import numpy as np
import warnings
import scipy

from util import data_validation 
from scipy.stats import vonmises, multivariate_normal, norm

from scipy.special import i0, i1

class VonMises(object):
    """
    VonMises class based on scipy.stats.vonmises
    """
    def __init__(self,**kwargs):
        super(VonMises,self).__init__()
        self.name   = 'vonmises'
        self.phi    = data_validation.angle(kwargs.pop('phi',0.))
        self.kappa  = kwargs.pop('kappa',None)
        self.R      = kwargs.pop('R',None)
        
        if (self.kappa is None) and (self.R is None):
            self.R = 0.0
        
        if self.kappa is None:
            self.calculate_kappa()  #uses self.R to calculate self.kappa
        else:
            self.calculate_R()      #uses self.kappa to calculate self.R      
            
        self.vm = vonmises(loc=self.phi,kappa=self.kappa)   #uses scipy.stats.vonmises
    
    def calculate_R(self):
        """
        Compute R based on kappa
        """
        self.R = self.A(self.kappa)
    
    def calculate_kappa(self):
        """
        Compute kappa based on R
        """
        if self.R>1:
            warnings.warn('Caution, R>1, R: '+str(self.R)+' R set to ~.99',Warning)
            self.R = .9999999999999999
        self.kappa = self.A_inv(self.R)
    
    def pdf(self,x):
        """
        computes von mises pdf
        """
        return self.vm.pdf(x)
    
    def random_samples(self,size):
        """
        Generates random samples of size = size
        """
        #return np.random.vonmises(self.phi,vm.kappa,size)
        return (self.vm.rvs(size=size)+np.pi)%(2*np.pi)-np.pi

    def convolute(self,vm2):
        '''
        Approximation of the convolution with another von mises
        Convolution of Von Mises disributions: Directional Statistics, Marida and Jupp page 45
        Numerical studies
        
        vm2:
            vonmises class with which convolution is performed
        '''
        assert isinstance(vm2,VonMises) # assert if the argument is a vonmises class
        #updates variables
        self.phi = self.phi+vm2.phi
        self.R  = self.R*vm2.R
        self.calculate_kappa()
        self.vm = scipy.stats.vonmises(loc=self.phi,kappa=self.kappa)

    def condition(self,vm2):
        '''
        Incorporating new measurements of the variable
        vm2:
            vonmises class with which conditioning is performed
        '''
        assert isinstance(vm2,VonMises) # assert if the argument is a vonmises class
        #updates variables
        c  = self.kappa/vm2.kappa
        x0 = c*np.cos(self.phi)
        y0 = c*np.sin(self.phi)
        x1 = np.cos(vm2.phi)
        y1 = np.sin(vm2.phi)

        self.phi = np.arctan2(y0+y1,x0+x1) 
        self.kappa = self.kappa+vm2.kappa
        self.calculate_R()
        self.vm = vonmises(loc=self.phi,kappa=self.kappa)

    def __str__(self):
        """
        Overloading __str__ to make print statement meaningful
        
        <format>
        Name                        :
        Parameters
            phi                     :
            kappa                   :
            R                       :
        """
        to_print = '{}  : {}\n'.format('Name'.ljust(34),self.name)
        to_print = to_print + '  {}  : {}\n'.format('phi'.ljust(32),str(self.phi))
        to_print = to_print + '  {}  : {}\n'.format('kappa'.ljust(32),str(self.kappa))
        to_print = to_print + '  {}  : {}'.format('R'.ljust(32),str(self.R))
        return to_print

    @staticmethod
    def A(kappa):
        '''R = A(kappa) = i1(kappa)/i0(kappa)'''
        return i1(kappa)/i0(kappa)
    
    @staticmethod
    def A_inv(R):
        '''A_inv(R) = kappa'''
        if R<0.53:
            kappa = 2*R+R**3+5/6.*R**5
        elif R<0.85:
            kappa = -0.4+1.39*R+0.43/(1-R)
        else:
            kappa = 1./(2*(1-R)-(1-R)**2-(1-R)**3)
        return kappa


class PN2(object):
    """
    Projected Normal with covariance equal to the 2nd order Identity matrix

    ux,uy,sd:
        
    samples:
    """
    def __init__(self,**kwargs):
        super(PN2,self).__init__()
        self.name   = 'Projected Normal'
        tmp_ux      = kwargs.pop('ux',None)
        tmp_uy      = kwargs.pop('uy',None)
        tmp_sd      = kwargs.pop('sd',None)
        samples     = kwargs.pop('samples',None)
 
        assert ((tmp_sd>0) or (tmp_sd is None))

        if any(k is None for k in (tmp_ux,tmp_uy,tmp_sd)):
            if samples is not None:
                assert samples.shape[1] == 2
                tmp_ux, tmp_uy = np.mean(samples,axis=0)
                tmp_sd = (np.sum(np.var(samples,axis=0))/2)**.5
    
        self.ux  = tmp_ux/tmp_sd
        self.uy  = tmp_uy/tmp_sd
        self._sd = tmp_sd
        
    def pdf(self,x):
        """
        Computes projected normal pdf
        """
        n2  = multivariate_normal(mean=np.zeros(2))
        D = self.ux*np.cos(x)+self.uy*np.sin(x)

        return (n2.pdf(np.asarray([self.ux,self.uy]))+D*norm.cdf(D)*norm.pdf((self.ux*np.sin(x)-self.uy*np.cos(x))))

    
    def equivalent_vonmises(self,n=1000,phiR=False):
        """
        Finds an equivalent the von mises pdf by calculating PN2 phi and R.
        phi and R are found by evaluating numericaly evaluating y = PN2.pdf(x) 
        and computing the resulting phi and R from y distribution.

        kwargs        
            n: The number of points used for the numerical calculation - default 1000 
            phiR: additionally output phi and R if flag is True - default False    
        """
        x = np.linspace(-np.pi,np.pi,n)
        y = self.pdf(x)

        R = np.sqrt((1./np.sum(y)**2)*(np.sum(y*np.cos(x))**2+np.sum(y*np.sin(x))**2))
        if not phiR:
            return VonMises(phi=np.arctan2(self.uy,self.ux),R=R)
        else:
            return VonMises(phi=np.arctan2(self.uy,self.ux),R=R),np.arctan2(self.uy,self.ux),R


    def __str__(self):
        """
        Overloading __str__ to make print statement meaningful
        
        <format>
        Name                        :
        Parameters
            phi                     :
            R                       :
        """
        tmp, phi, R = self.equivalent_vonmises(phiR=True)
        to_print = '{}  : {}\n'.format('Name'.ljust(34),self.name)
        to_print = to_print + '  {}  : {}\n'.format('ux'.ljust(32),str(self.ux*self._sd))
        to_print = to_print + '  {}  : {}\n'.format('uy'.ljust(32),str(self.uy*self._sd))
        to_print = to_print + '  {}  : {}\n'.format('sd'.ljust(32),str(self._sd))
        to_print = to_print + '  {}  : {}\n'.format('phi'.ljust(32),str(phi))
        to_print = to_print + '  {}  : {}'.format('R'.ljust(32),str(R))
        return to_print








