import numpy as np
from .directionalstats import VonMises, PN2
from ..sampling.sampling import resampling_wheel
from ..sensormodel.sensormodel import SensorModel
from ..motionmodel.motionmodel import MotionModel
#from sklearn.neighbors.kde import KernelDensity as gaussian_kde
from scipy.stats import gaussian_kde

import time


class kde_0():
    @staticmethod
    def evaluate(x):
        return np.ones(x.shape[1])

class DualMCL(object):
    """
    Base class for DualMCL.
    """
    def __init__(self,sensormodel,motionmodel,**kwargs):
        super(DualMCL,self).__init__()
        self.name = 'MCLcore'
        assert isinstance(sensormodel,SensorModel)
        assert isinstance(motionmodel,MotionModel)
    
        self.sensormodel    = sensormodel
        self.motionmodel    = motionmodel

        self.verbose        = kwargs.get('verbose',False)
        self.nsamples       = kwargs.get('nsamples',800)
        self.orientsamples  = kwargs.get('orientsamples',10)
        self.particles_dim  = kwargs.get('particles_dim',3)
        
        #current state
        self.particles      = kwargs.get('particles',
                                            np.random.rand(self.nsamples,self.particles_dim))
        self.kde            = kwargs.get('kde',kde_0())
        self.vm             = kwargs.get('vm',VonMises(phi=0.,R=.5))
        self.odometry_last  = kwargs.get('odometry',np.asarray([0.,0.,0.]))
        self.x              = kwargs.get('x',0.)
        self.y              = kwargs.get('y',0.)
        self.v              = kwargs.get('v',1000.)
        
        #cumulative
        self.log = {'errors':[],'pos_estimates':[],'particles':[],'sensor_samples':[],
                        'orientation_samples':[]}

    def run(self,odometry,measurements,xtrue=None,save_log=False):
        """
        runs dual MCL over motion actions and sensor measurements
        """
        odometry_prev   = self.odometry_last
        x_prev          = self.x
        y_prev          = self.y
        v_prev          = self.v
        kde_prev        = self.kde        
        vm              = self.vm

        for i in range(odometry.shape[0]):
            t1 = time.time()
            odometry_next       = odometry[i,:] 
            measurement_next    = measurements[i:i+1,:]

            ########## Von mises orientation update from odometry ##########
            phi_dif     = odometry_next[2] - odometry_prev[2]
            trans_dif   = ((odometry_next[0]-odometry_prev[0])**2 
                                +(odometry_next[1]-odometry_prev[1])**2)**.5
            vm_update   = VonMises(phi=phi_dif,R=1-trans_dif*0.0025/.75) #odometry noise
            vm.convolute(vm_update)  

            ########## Sampling from sensor model and Von mises ##########
            #Sample sensor model
            samples_2d  = self.sensormodel.sample((measurement_next,measurement_next), 
                                                                    nsamples=self.nsamples)
            #Sample orientation
            samples_or  = vm.random_samples(self.nsamples*self.orientsamples)  
            
            particles   = np.ones((self.orientsamples,1,1))*samples_2d[None,:,:]
            particles   = np.reshape(particles,(self.nsamples*self.orientsamples,2))
            particles   = np.concatenate((particles,samples_or[:,None]),axis=1) #3D particles

            ########## Resampling at t-1  ##########
            #Move particles backwards to t-1
            p_prev      = self.motionmodel.backward(particles,(odometry_prev,odometry_next))
            w_prev      = kde_prev.evaluate(p_prev[:,0:2].T)            
            #w_prev      = np.exp(kde_prev.score_samples(p_prev[:,0:2])) #Weights at t-1
            
            #TODO: Check condition for rebooting MCL later             
            if np.max(w_prev)<1e-10:
                print 'kde does not match samples from GP'
                w_prev  = np.ones_like(w_prev) #Reboot kde
                vm_d    = VonMises(phi=0.,R=0.01)
                self.vm.convolute(vm_d)   #Reboot orientation
                v_prev  = 1000000. # Makes this measurement not count
    
            p_prev = resampling_wheel(p_prev,w_prev,Nsamples=self.nsamples)

            ########## Moving to current time and updating kde ##########            
            #move particle forwards to t
            p_new       = self.motionmodel.forward(p_prev,(odometry_prev,odometry_next))
            # new kde                    
            bandwidth   = p_new.shape[0]**(-1./6)
            b_std       = np.std(p_new,axis=0)
            b_std       = (b_std[0]*b_std[1])**.5
            bandwidth   = bandwidth*b_std
            #kde         = gaussian_kde(kernel='linear',bandwidth=bandwidth*2.5,
            #                             algorithm='kd_tree',rtol=1e-2)
            #kde.fit(p_new[:,0:2])
            kde = gaussian_kde(p_new[:,0:2].T)            
            w_new       = kde.evaluate(p_new[:,0:2].T)            
            #w_new       = np.exp(kde.score_samples(p_new[:,0:2]))           

            pos_estimate= np.sum(np.asarray([w_new,w_new]).T*p_new[:,0:2],axis=0)/np.sum(w_new)
    
            ########## Von Mises orientation update from measurements ##########
            x  = pos_estimate[0]#np.mean(p_new[:,0])
            y  = pos_estimate[1]#np.mean(p_new[:,1])
            v  = np.var(p_new[:,0])+np.var(p_new[:,1])
        
            ux = x-x_prev
            uy = y-y_prev
            sd = (v+v_prev)**.5

            pn = PN2(ux=ux,uy=uy,sd=sd)
            vm_update = pn.equivalent_vonmises()
            vm.condition(vm_update)
        
            ########## Replacing 5% of particle by 'random' ones from sampling ##########
            index = np.random.randint(self.nsamples,size=np.round(self.nsamples*.05))
            p_new[index,:] = particles[index,:]

            ########## Updating ##########
            odometry_prev = odometry_next    
            kde_prev = kde    
            x_prev = x
            y_prev = y
            v_prev = v
            
            t2 = time.time()
            
            ########## Savin and plotting information ##########
            if xtrue is not None:
                error       = xtrue[i,:]-pos_estimate
                mse_error   = np.sqrt(np.sum(error*error))
            
            if self.verbose:
                print '  {} : Error: {}      Computation time: {}'.format(str(i).ljust(4),
                                                    str(mse_error).ljust(8),str(t2-t1).ljust(8))
            
            if save_log:
                if xtrue is not None:
                    self.log['errors'].append(mse_error)
                self.log['pos_estimates'].append(pos_estimate)
                self.log['particles'].append(p_new)
                self.log['sensor_samples'].append(samples_2d)
                self.log['orientation_samples'].append(samples_or)                

        self.odometry_last = odometry_next
        self.x = x
        self.y = y
        self.v = v
        self.vm = vm

    def __str__(self):
        """
        Overloading __str__ to make print statement meaningful
        
        <format>
        Name                        :
        Number of samples           :
        Sensor model                :
        Motion model                :
        """
        to_print = '{}  : {}\n'.format('Name'.ljust(34),self.name)
        to_print = to_print + '{}  : {}\n'.format('Name'.ljust(34),self.pathloss.name)
        to_print = to_print + '{}  : {}\n'.format('Number of samples'.ljust(34),self.nsamples)
        to_print = to_print + '{}  : {}\n'.format('Sensor model'.ljust(34),self.sensor_model.name)
        to_print = to_print + '{}  : {}\n'.format('Motion model'.ljust(34),self.motion_model.name)
        return to_print

    
    


