import numpy as np
import scipy


class Likelihood(object):
    """
    Base class for likelihoods. Each inherited classs needs to implement jointpdf
    """
    def __init__(self,**kwargs):
        super(Likelihood,self).__init__()

    def jointpdf(self,Xn,Yn,Varn):
        """
        Computes de loglikelihood of measurements Yn with variance Varn to have been taken 
        at positions Xn.
        """
        raise NotImplementedError 

class Gaussian(Likelihood):
    """
    Gaussian likelihood of all accespoints calculated as the geometric mean of each one.
    
    kwargs  
        deltax: see prob()
        confidence_zero: 
        pzero: 
    """
    def __init__(self,**kwargs):
        super(Gaussian,self).__init__()
        self.deltax             = kwargs.pop('deltax',.025)
        self.confidence_zero    = kwargs.pop('confidence_zero',.3)
        self.pzero              = kwargs.pop('pzero',.05)

    def prob(self,x,mu,var,**kwargs):
        """
        Calculates the probability of each individual access point by integrating the density
        function of N(mu,var) from x-deltax to x+deltax.

        x: 
            point which probability is calculated
        mu, var:
            mean and variance of the gausian density
        kwargs    
            deltax: variation around x where the integration is performed - default .15

        <math>    
            cdf(x) = 0.5*(1+erf((x-mu)/sqrt(2var)))
            prob = cdf(x+deltax)-cdf(x-deltax)
        """

        deltax = self.deltax

        xplus  = (x+self.deltax-mu)/(var*2)**0.5
        xless  = (x-self.deltax-mu)/(var*2)**0.5
        px     = 0.5*(scipy.special.erf(xplus)-scipy.special.erf(xless))
        return np.clip(px,1e-15,np.inf)
    
    def jointpdf(self,model,Xn,Yn,Varn,**kwargs):
        """
        model:
            mapping class that generates the predictions for Xn
        Xn:
            tp test points as a numpy array [tp, 2]
        Yn: 
            mp measurement points of all n wireless measurements as a numpy array [mp, n]
        Varn: 
            mp var measurement points of all n wireless measurements as a numpy array [mp, n]
        kwargs:
            log: if True, ouputs the log joint probability instead - default False
            
        """
        logFlag         = kwargs.get('log',False)
        deltax          = self.deltax
        confidence_zero = self.confidence_zero
        pzero           = self.pzero

        if Yn.ndim==1:
            measurement = Yn.copy().reshape(1,len(Yn))
            #measure_var = Varn.copy().reshape(1,len(Varn))
        else:
            measurement = Yn.copy()
            #measure_var = Varn.copy()
        
        ypredict, varpredict = model.predict(Xn)       
        assert measurement.shape[1]==ypredict.shape[1]
        
        n_testp  = measurement.shape[0]
        n_points = ypredict.shape[0]
        n_ap     = measurement.shape[1]
    
        log_pXV  = np.zeros((n_points,n_testp))
    
        if varpredict.shape[1] == 1:
            var  = varpredict.copy()
            
        for i in range(0,n_testp):
            xs   = np.reshape(measurement[i,:],(1,n_ap))
            mus  = ypredict
            if varpredict.shape[1] > 1:
                var  = varpredict
        
            px      = (1-self.pzero)*self.prob(xs,mus,var,deltax=deltax)
            px_zero = np.zeros_like(xs)         # Probability if the measurement is zero
            px_zero[np.where(xs==0)] = pzero  
            px      = px + px_zero              #Total probability 
        
            #Zero measurements are not trusted the same necessarily 
            # - if confidence_zero = 1 they are
            #cz = confidence_zero
            if confidence_zero == 0:            
                cz = (1.*(n_ap-np.sum(measurement[i,:]==0)))/(np.sum(measurement[i,:]==0)*1.)
            else:
                if confidence_zero is None:            
                    cz = 1./(np.sum(measurement[i,:]==0)*1.)
                else:
                    cz = confidence_zero
            alpha = np.ones(n_ap)
            alpha[np.where(measurement[i,:]==0)] = cz   #all zeros get one vote
           
            alpha = alpha/np.sum(alpha)     #normalize
            log_px = np.sum(alpha*np.log(px),axis=1,keepdims=True)
            log_pXV[:,i] = log_px.flatten()

        if logFlag:
            return np.clip(log_pXV,-200,200)
        else:
            return np.exp(log_pXV)
